/*
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <bluetooth/services/nus.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include "AFE4900.h"
#include "typedef.h"
#include "bt_nus.h"
#include "data.h"
#include "datetime.h"
#include "i2c.h"
#include "error.h"
#include "flash_thread.h"
#include "localtime.h"

#define BT_NUS_CMD_CAPTURE_LEN   7
#define BT_NUS_CMD_MEASURE_LEN   7
#define BT_NUS_CMD_READ_LEN      4
#define BT_NUS_CMD_SKIN_LEN      4
#define BT_NUS_CMD_STOP_LEN      4
#define BT_NUS_CMD_TIME_LEN      DATE_TIME_BUF_LEN
#define BT_NUS_MAX_DATA_LEN      1024
#define BT_NUS_RX_BUF_SIZE       256
#define BT_NUS_TX_BUF_SIZE       1024

typedef enum {
    BT_NUS_CMD_UNKNOWN = 0, 
    BT_NUS_CMD_TIME,
    BT_NUS_CMD_SKIN,  
    BT_NUS_CMD_CAPTURE,
    BT_NUS_CMD_MEASURE,
    BT_NUS_CMD_STOP, 
    BT_NUS_CMD_READ,
} bt_nus_cmd_t;

static bool            _gBtNusTxCompleted = false;
static bool            _gBtNusTxBufFull   = false;
static bt_nus_state_t  _gBtNusState = BT_NUS_STATE_DISCONNECT;
static uint8_t         _gBtNusTxBuf[BT_NUS_TX_BUF_SIZE];
static uint16_t        _gBtNusTxBufHeader  = 0;
static uint16_t        _gBtNusTxBufTail    = 0;
static uint32_t        _gBtNusMtu    = 0;
static uint32_t        _gBtNusSentBytes = 0;

static void _btNusReceiveCb(struct bt_conn *conn, const uint8_t* const pData, const uint16_t len);
static void _btNusSentCb(struct bt_conn *conn);

static struct bt_nus_cb _gNusCb = {
    .received = _btNusReceiveCb,
    .sent     = _btNusSentCb,
};

static bt_nus_cmd_t _btNusGetCmd( const uint8_t* const pData, const uint16_t len) {
    bt_nus_cmd_t cmd = BT_NUS_CMD_UNKNOWN;
    if ((len == BT_NUS_CMD_CAPTURE_LEN) &&
        (pData[0] == 0x43) && //  "CAPTURE"
        (pData[1] == 0x41) &&
        (pData[2] == 0x50) &&
        (pData[3] == 0x54) &&
        (pData[4] == 0x55) &&
        (pData[5] == 0x52) &&
        (pData[6] == 0x45)) {
        cmd = BT_NUS_CMD_CAPTURE;
    } else if ((len == BT_NUS_CMD_MEASURE_LEN) &&
        (pData[0] == 0x4D) && //  "MEASURE"
        (pData[1] == 0x45) &&
        (pData[2] == 0x41) &&
        (pData[3] == 0x53) &&
        (pData[4] == 0x55) &&
        (pData[5] == 0x52) &&
        (pData[6] == 0x45)) {
        cmd = BT_NUS_CMD_MEASURE;
    } else if ((len == BT_NUS_CMD_READ_LEN) &&
        (pData[0] == 0x52) && //  "READ"
        (pData[1] == 0x45) &&
        (pData[2] == 0x41) &&
        (pData[3] == 0x44)) {
        cmd = BT_NUS_CMD_READ;
    } else if ((len == BT_NUS_CMD_STOP_LEN) &&
        (pData[0] == 0x53) && //  "STOP"
        (pData[1] == 0x54) &&
        (pData[2] == 0x4F) &&
        (pData[3] == 0x50)) {
        cmd = BT_NUS_CMD_STOP;
    } else if ((len == BT_NUS_CMD_TIME_LEN) &&
        (pData[0] == 0x54) && //  "TIME"
        (pData[1] == 0x49) &&
        (pData[2] == 0x4D) &&
        (pData[3] == 0x45)) {
        cmd = BT_NUS_CMD_TIME;
    }

    return cmd;
}

bool _btNusIsTxBufEmpty(void) {
    if (!_gBtNusTxBufFull && _gBtNusTxBufHeader == _gBtNusTxBufTail) {
        return true;
    }

    return false;
}


void btNusTxBufClear(void) {
    _gBtNusTxBufHeader = 0;
    _gBtNusTxBufTail   = 0;
    _gBtNusTxBufFull   = false;
}

bool btNusTxBufEmpty(void) {
    if ((!_gBtNusTxBufFull) && (_gBtNusTxBufHeader == _gBtNusTxBufTail)) {
        return true;
    }
    
    return false;
}

uint16_t _btNusTxBufRead(uint8_t* buf, const int len) {
    int i;
    int count = 0;
    
    for(i=0;i<len;i++) {
        if (_gBtNusTxBufFull || _gBtNusTxBufHeader != _gBtNusTxBufTail) {
            buf[i] = _gBtNusTxBuf[_gBtNusTxBufTail];
            _gBtNusTxBufTail++;
            if (_gBtNusTxBufTail >= BT_NUS_TX_BUF_SIZE) {
                _gBtNusTxBufTail = 0;
            }
            _gBtNusTxBufFull = false;
            count++;
        } else {
            break;
        }
    }
    
    return count;
}

uint16_t _btNusTxBufWrite(const uint8_t *buff, const int len) {
    int i;
    uint16_t count = 0;
    if (!_gBtNusTxBufFull) {
        for(i=0;i<len;i++) {
            _gBtNusTxBuf[_gBtNusTxBufHeader] = buff[i];
            _gBtNusTxBufHeader++;
            if (_gBtNusTxBufHeader >= BT_NUS_TX_BUF_SIZE) {
                _gBtNusTxBufHeader = 0;
            }

            count++;
            if (_gBtNusTxBufHeader == _gBtNusTxBufTail) {
                printk("_btNusWriteTxBuf Full\n");
                _gBtNusTxBufFull = true;
                break;
            }
        }
    }
    
    return count;
}

uint32_t btNusSend(void) {
    printk("btNusSend() e\n");
    static uint8_t sbuf[BT_NUS_MAX_DATA_LEN];
    uint16_t len;
    int error = 0;

    len = _btNusTxBufRead(sbuf, _gBtNusMtu);

    if (len > 0) {
        printk("len=%u\n", len);
        _gBtNusSentBytes += len;
        printk("bt_nus_send(%u)\n", len);
        error = bt_nus_send(NULL, sbuf, len);
    }

    return error;
}

static void _btNusReceiveCb(struct bt_conn *conn, const uint8_t* const pData, const uint16_t len) {
    printk("_btReceiveCb(), le=%u e\n", len);
    dataPrintChar("nusr", len, pData);
    dataPrintHex("nusr", len, pData);
    
    const bt_nus_cmd_t cmd = _btNusGetCmd(pData, len); 
     if (cmd == BT_NUS_CMD_CAPTURE) {
        printk("CAPTURE\n");
        btNusSetState(BT_NUS_STATE_CAPTURE);
    } else if (cmd == BT_NUS_CMD_MEASURE) {
        printk("MEASURE\n");
        btNusSetState(BT_NUS_STATE_MEASURE);
        flashThreadCreate();
    } else if (cmd == BT_NUS_CMD_READ) {
        printk("READ\n");
        btNusSetState(BT_NUS_STATE_READ);
    } else if (cmd == BT_NUS_CMD_STOP) {
        printk("STOP\n");
        btNusSetState(BT_NUS_STATE_STOP_BEGIN);
        flashThreadAbort();
    }  else if (cmd == BT_NUS_CMD_TIME) {
            printk("TIME\n");
            btNusSetState(BT_NUS_STATE_SYNC);
            datetime_t dateTime;
            data2Datetime(&pData[0], &dateTime);
            datetimePrint("datetime sync", &dateTime);
            ltSetDatetime(&dateTime);
    }
}

static void _btNusSentCb(struct bt_conn *conn) {
}

int btNusInit(void) {
    const int ret = bt_nus_init(&_gNusCb);
    if (ret < 0) {
        printk("bt_nus_init() failed, err=0x%x\n", ret);
        return ERROR_NUS_INITIALIZE_FAIL;
    }

    return 0;
}

uint16_t btNusWriteData(uint8_t *pValue, uint16_t len) {
    const uint16_t ret = _btNusTxBufWrite(pValue, len);
    _gBtNusTxCompleted = true;

    return ret;
}

uint32_t btNusGetMtu(void) {
    return _gBtNusMtu;
}

void btNusSetMtu(const uint32_t mtu) {
    _gBtNusMtu = mtu;
}

bt_nus_state_t btNusGetState(void) {
    return _gBtNusState;
}

void btNusSetState(const bt_nus_state_t state) {
    _gBtNusState = state;
    btNusPrintState();
}

bool btNusGetTxCompleted(void) {
    return _gBtNusTxCompleted;
}

void btNusSetTxCompleted(const bool completed) {
    _gBtNusTxCompleted = completed;
}

void btNusPrintState(void) {
    const bt_nus_state_t state = btNusGetState();
    switch (state) {
        case BT_NUS_STATE_DISCONNECT:
        {
            printk("state = disconnect\n"); 
        }
        break;

        case BT_NUS_STATE_CONNECT:
        {
            printk("state = connect\n"); 
        }
        break;
        
        case BT_NUS_STATE_SYNC:
        {
            printk("state = sync\n"); 
        }
        break;
        
        case BT_NUS_STATE_SKIN:
        {
            printk("state = skin\n"); 
        }
        break;
        
        case BT_NUS_STATE_CAPTURE:
        {
            printk("state = capture\n"); 
        }
        break;
        
        case BT_NUS_STATE_MEASURE_BEGIN:
        {
            printk("state = measure begin\n"); 
        }
        break;
        
        case BT_NUS_STATE_MEASURE:
        {
            printk("state = measure\n");
        }
        break;
        
        case BT_NUS_STATE_MEASURE_ENDING:
        {
            printk("state = measure ending\n");
        }
        break;
        
        case BT_NUS_STATE_MEASURE_ENDED:
        {
            printk("state = measure ended\n");
        }
        break;
        
        case BT_NUS_STATE_STOP_BEGIN:
        {
            printk("state = stop begin\n");
        }
        break;
        
        case BT_NUS_STATE_STOP:
        {
            printk("state = stop\n");
        }
        break;
        
        case BT_NUS_STATE_READ_BEGIN:
        {
            printk("state = read begin\n");
        }
        break;
        
        case BT_NUS_STATE_READ:
        {
            printk("state = read\n");
        }
        break;
    }    
}
