/*
 * Copyright (c) 2016 Intel Corporation.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <inttypes.h>
#include <jesd216.h>
#include <stdio.h>
#include <string.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/flash.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/sys/printk.h>
#include "error.h"
#include "flash_nor.h"

#if DT_HAS_COMPAT_STATUS_OKAY(jedec_spi_nor)
#define FLASH_NODE DT_COMPAT_GET_ANY_STATUS_OKAY(jedec_spi_nor)
#elif DT_HAS_COMPAT_STATUS_OKAY(nordic_qspi_nor)
#define FLASH_NODE DT_COMPAT_GET_ANY_STATUS_OKAY(nordic_qspi_nor)
#elif DT_HAS_COMPAT_STATUS_OKAY(st_stm32_qspi_nor)
#define FLASH_NODE DT_COMPAT_GET_ANY_STATUS_OKAY(st_stm32_qspi_nor)
#elif DT_HAS_COMPAT_STATUS_OKAY(st_stm32_ospi_nor)
#define FLASH_NODE DT_COMPAT_GET_ANY_STATUS_OKAY(st_stm32_ospi_nor)
#else
#error Unsupported flash driver
#define FLASH_NODE DT_INVALID_NODE
#endif

/* Instruction codes */
#define BLS_CODE_PROGRAM          0x02 /**< Page Program */
#define BLS_CODE_PROGRAM_4B       0x12 /**< Page Program with 4-Byte Address */
#define BLS_CODE_READ             0x03 /**< Read Data */
#define BLS_CODE_READ_4B          0x13 /**< Read Data with 4-Byte Address */
#define BLS_CODE_READ_STATUS      0x05 /**< Read Status Register */
#define BLS_CODE_WRITE_ENABLE     0x06 /**< Write Enable */
#define BLS_CODE_SECTOR_ERASE     0x20 /**< Sector (4K-bytes) Erase */
#define BLS_CODE_SECTOR_ERASE_4B  0x21 /**< Sector (4K-bytes) Erase with 4-Byte Address */
#define BLS_CODE_MDID             0x90 /**< Manufacturer Device ID */
#define BLS_CODE_JEDEC_ID         0x9F /**< JEDEC ID */

#define BLS_CODE_DP               0xB9 /**< Power down */
#define BLS_CODE_RDP              0xAB /**< Power standby */

/* Erase instructions */
#define BLS_CODE_ERASE_4K         0x20 /**< Sector Erase */
#define BLS_CODE_ERASE_32K        0x52
#define BLS_CODE_ERASE_64K        0xD8
#define BLS_CODE_ERASE_ALL        0xC7 /**< Mass Erase */

/* Bitmasks of the status register */
#define BLS_STATUS_SRWD_BM        0x80
#define BLS_STATUS_BP_BM          0x0C
#define BLS_STATUS_WEL_BM         0x02
#define BLS_STATUS_WIP_BM         0x01

/* Status register */
#define BLS_STATUS_BIT_BUSY       0x01 /* Busy bit */
#define BLS_STATUS_BIT_WEL        0x02 /* Write enable latch bit */
#define BLS_STATUS_BIT_BP0        0x04 /* Block protect bit */ 
#define BLS_STATUS_BIT_BP1        0x08 /* Block protect bit */ 
#define BLS_STATUS_BIT_BP2        0x10 /* Block protect bit */ 
#define BLS_STATUS_BIT_TB         0x20 /* Top/Bottom protect bit */ 
#define BLS_STATUS_BIT_SEC        0x40 /* Sector protect bit */ 
#define BLS_STATUS_BIT_SRP        0x80 /* Status register protect bit */ 

/* Part specific constants */
#define BLS_PROGRAM_PAGE_SIZE     256
#define BLS_ERASE_SECTOR_SIZE     4096
#define EXT_FLASH_PAGE_SIZE       4096

#define EONMID                    0x1C
#define MXMID                     0xC2
#define WINBONDMID                0xEF

/* JEDEC ID */
#define JEDEC_ID_UNKNOWN          0xFFFFFF
#define JEDEC_ID_W25Q32JV         0xEF4016
#define JEDEC_ID_W25Q256JV        0xEF4019

/* Density */
#define DENSITY_32b               0x400000
#define DENSITY_128b              0x1000000
#define DENSITY_256b              0x2000000

const struct device* const _gFlashDev           = DEVICE_DT_GET(FLASH_NODE);
static bool   _gFlashNorInitialzed = false;
static int    _gJedecId = JEDEC_ID_UNKNOWN;
static off_t  _gErasedOffset       = 0;
static off_t  _gReadOffset         = 0;
static off_t  _gWrittenOffset      = 0;
static size_t _gDensity            = 0;
//static size_t _gReadLen            = 0;
//static size_t _gWrittenLen         = 0;

static int _flashNorGetSectorNum(void);

int _flashNorCalcDensity(void) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }

    const unsigned int sectorNum = _flashNorGetSectorNum();
    _gDensity = sectorNum * BLS_ERASE_SECTOR_SIZE;
    printk("_gDensity=%d\n", _gDensity);

    return 0;
}

static int _flashNorGetSectorNum(void) {
    int sectorNum;
    switch (_gJedecId) {
        case JEDEC_ID_W25Q256JV:
            sectorNum = 8192;
            break;

        case JEDEC_ID_W25Q32JV:
        default:
            sectorNum = 1024;
    }

    return sectorNum;
}

static bool _flashNorIsEraseNeeded(const size_t offset) {
    if (offset < _gErasedOffset) {
        return false;
    }

    return true;
}

int _flashNorReadData(const off_t offset, const size_t len, uint8_t *buf) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }

    const int ret = flash_read(_gFlashDev, offset, buf, len);
    if (ret < 0) {
        printk("flash_read(%s, %ld, %p, %u) faild 0x%x\n", _gFlashDev->name, offset, buf, len, ret);
        return ERROR_FSTORAGE_READ_FAIL;
    
    }

    return 0;
}

int _flashNorReadJedecId(void) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }

    uint8_t id[3];

    const int ret = flash_read_jedec_id(_gFlashDev, id);
    if (ret < 0) {
        printk("JEDEC ID read failed: %d\n", ret);
        return ERROR_FSTORAGE_READ_FAIL;
    }

    printk("jedec-id = [%02x %02x %02x]\n", id[0], id[1], id[2]);
    if ((id[0] == 0xEF) && (id[1] == 0x40) && (id[2] == 0x16)) {
        _gJedecId = JEDEC_ID_W25Q32JV;
    } else if ((id[0] == 0xEF) && (id[1] == 0x40) && (id[2] == 0x19)) {
        _gJedecId = JEDEC_ID_W25Q256JV;
    } else {
        _gJedecId = JEDEC_ID_UNKNOWN;
    }

    return 0;
}

int _flashNorWriteData(off_t offset, const size_t len, uint8_t* buf) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }

    const int ret = flash_write(_gFlashDev, offset, buf, len);
    if (ret < 0) {
        printk("flash_write(%s, %ld, %p, %u) faild 0x%x\n", _gFlashDev->name, offset, buf, len, ret);
        return ERROR_FSTORAGE_WRITE_FAIL;
    }

    return 0;
}

int flashNorInit(void) {
    if (!device_is_ready(_gFlashDev)) {
        printk("%s: device not ready\n", _gFlashDev->name);
        return -1;
    }

    _gFlashNorInitialzed = true;

    _flashNorReadJedecId();
    _flashNorCalcDensity();

#if 0
    printk("erasing\n");
    flashNorEraseData(0, BLS_ERASE_SECTOR_SIZE);
    uint8_t rbuf[BLS_PROGRAM_PAGE_SIZE];
    uint8_t wbuf[BLS_PROGRAM_PAGE_SIZE];
    flashNorReadData(0, BLS_PROGRAM_PAGE_SIZE, rbuf);
    for (int i=0; i<10; i++) {
        printk("rbuf[%d]=0x%x\n", i, rbuf[i]);
    }

    for (int i=0; i<BLS_PROGRAM_PAGE_SIZE; i++) {
        wbuf[i] = i;
    }

    printk("writing\n");
    flashNorWriteData(0, BLS_PROGRAM_PAGE_SIZE, wbuf);
    flashNorReadData(0, BLS_PROGRAM_PAGE_SIZE, rbuf);
    for (int i=0; i<10; i++) {
        printk("rbuf[%d]=0x%x\n", i, rbuf[i]);
    }
#endif

    return 0;
}

int flashNorEraseData(const off_t offset, const size_t len) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }

    const int ret = flash_erase(_gFlashDev, offset, len);
    if (ret < 0) {
        printk("flash_erase(%s, %ld, %u) faild %d\n", _gFlashDev->name, offset, len, ret);
        return ERROR_FSTORAGE_ERASE_FAIL;
    }

    return 0;
}

int flashNorReadData(const off_t offset, const size_t len, uint8_t *buf) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }

    if (_flashNorReadData(offset, len, buf) != 0) {
        printk("extFlashRead() failed");
        return ERROR_NOR_READ_FAIL;
    }

    return 0;
}

int flashNorWriteData(const size_t len, uint8_t* buf) {
    if (!flashNorGetInitialzed()) {
        return ERROR_NOR_UNINITIALIZED;
    }  

    const size_t density = flashNorGetDensity();
    const size_t woffset  = _gWrittenOffset + len;
    if (woffset > density) {
        printk("flashNorWriteData(): end of NOR woffset=0x%x, len=0x%x, density=0x%x\n", woffset, len, density);
        return ERROR_NOR_REACH_END;
    }    

    if (_flashNorIsEraseNeeded(woffset)) {
        if (flashNorEraseData(_gErasedOffset, BLS_ERASE_SECTOR_SIZE) != 0) {
            printk("extFlashErase failed\n");
            return ERROR_NOR_ERASE_FAIL;
        }
        _gErasedOffset += BLS_ERASE_SECTOR_SIZE;
    }

    if (_flashNorWriteData(_gWrittenOffset, len, buf) != 0)  {
        printk("extFlashWrite failed\n");
        return ERROR_NOR_WRITE_FAIL;
    }

    _gWrittenOffset +=  len;

    return 0;
}   


size_t flashNorGetDensity(void) {
    return _gDensity;
}

size_t flashNorGetEraseSectorSize(void) {
    return BLS_ERASE_SECTOR_SIZE;
}

bool flashNorGetInitialzed(void) {
    return _gFlashNorInitialzed;
}

size_t flashNorGetReadOffset(void) {
    return _gReadOffset;
}

void flashNorSetReadOffset(const size_t offset) {
    _gReadOffset = offset;
}

size_t flashNorGetWrittenOffset(void) {
    return _gWrittenOffset;
}

void flashNorSetWrittenOffset(const size_t offset) {
    _gWrittenOffset = offset;
}



